@@ -3,19 +3,29 @@
 package KiokuDB::Backend::DBI;
 use Moose;
 
-use MooseX::Types -declare => [qw(ValidColumnName)];
+use Moose::Util::TypeConstraints;
 
-use MooseX::Types::Moose qw(ArrayRef HashRef Str);
+use MooseX::Types -declare => [qw(ValidColumnName SchemaProto)];
+
+use MooseX::Types::Moose qw(ArrayRef HashRef Str Defined);
+
+use Moose::Util::TypeConstraints qw(enum);
 
 use Data::Stream::Bulk::DBI;
+use SQL::Abstract;
+use JSON;
+use Scalar::Util qw(weaken);
 
 use KiokuDB::Backend::DBI::Schema;
-
-use SQL::Abstract;
+use KiokuDB::TypeMap;
+use KiokuDB::TypeMap::Entry::DBIC::Row;
+use KiokuDB::TypeMap::Entry::DBIC::ResultSourceHandle;
+use KiokuDB::TypeMap::Entry::DBIC::ResultSet;
+use KiokuDB::TypeMap::Entry::DBIC::Schema;
 
 use namespace::clean -except => 'meta';
 
-our $VERSION = "0.08";
+our $VERSION = "1.11_01";
 
 with qw(
     KiokuDB::Backend
@@ -36,6 +46,7 @@ my @reserved_cols = ( @std_cols, 'data' );
 my %reserved_cols = ( map { $_ => 1 } @reserved_cols );
 
 subtype ValidColumnName, as Str, where { not exists $reserved_cols{$_} };
+subtype SchemaProto, as Defined, where { !ref($_) || blessed($_) and $_->isa("DBIx::Class::Schema::KiokuDB") };
 
 sub new_from_dsn {
     my ( $self, $dsn, @args ) = @_;
@@ -54,6 +65,12 @@ sub BUILD {
 
 has '+serializer' => ( default => "json" ); # to make dumps readable
 
+has json => (
+    isa => "Object",
+    is  => "ro",
+    default => sub { JSON->new },
+);
+
 has create => (
     isa => "Bool",
     is  => "ro",
@@ -65,7 +82,6 @@ has 'dsn' => (
     is  => "ro",
 );
 
-
 has [qw(user password)] => (
     isa => "Str",
     is  => "ro",
@@ -76,6 +92,66 @@ has dbi_attrs => (
     is  => "ro",
 );
 
+has mysql_strict => (
+    isa => "Bool",
+    is  => "ro",
+    default => 1,
+);
+
+has sqlite_sync_mode => (
+    isa => enum([qw(0 1 2 OFF NORMAL FULL off normal full)]),
+    is  => "ro",
+    predicate => "has_sqlite_fsync_mode",
+);
+
+has on_connect_call => (
+    isa => "ArrayRef",
+    is  => "ro",
+    lazy_build => 1,
+);
+
+sub _build_on_connect_call {
+    my $self = shift;
+
+    my @call;
+
+    if ( $self->mysql_strict ) {
+        push @call, sub {
+            my $storage = shift;
+
+            if ( $storage->can("connect_call_set_strict_mode") ) {
+                $storage->connect_call_set_strict_mode;
+            }
+        };
+    };
+
+    if ( $self->has_sqlite_fsync_mode ) {
+        push @call, sub {
+            my $storage = shift;
+
+            if ( $storage->sqlt_type eq 'SQLite' ) {
+                $storage->dbh_do(sub { $_[1]->do("PRAGMA synchronous=" . $self->sqlite_sync_mode) });
+            }
+        };
+    }
+
+    return \@call;
+}
+
+has dbic_attrs => (
+    isa => "HashRef",
+    is  => "ro",
+    lazy_build => 1,
+);
+
+sub _build_dbic_attrs {
+    my $self = shift;
+
+    return {
+        on_connect_call => $self->on_connect_call,
+    };
+}
+
 has connect_info => (
     isa => ArrayRef,
     is  => "ro",
@@ -85,14 +161,20 @@ has connect_info => (
 sub _build_connect_info {
     my $self = shift;
 
-    return [ $self->dsn, $self->user, $self->password, $self->dbi_attrs ];
+    return [ $self->dsn, $self->user, $self->password, $self->dbi_attrs, $self->dbic_attrs ];
 }
 
 has schema => (
     isa => "DBIx::Class::Schema",
     is  => "ro",
     lazy_build => 1,
-    handles => [qw(deploy)],
+    handles => [qw(deploy kiokudb_entries_source_name)],
+);
+
+has schema_proto => (
+    isa => SchemaProto,
+    is  => "ro",
+    default => "KiokuDB::Backend::DBI::Schema",
 );
 
 has schema_hook => (
@@ -104,9 +186,11 @@ has schema_hook => (
 sub _build_schema {
     my $self = shift;
 
-    my $schema = KiokuDB::Backend::DBI::Schema->clone;
+    my $schema = $self->schema_proto->clone;
 
-    $schema->source("entries")->add_columns(@{ $self->columns });
+    unless ( $schema->kiokudb_entries_source_name ) {
+        $schema->define_kiokudb_schema( extra_entries_columns => $self->columns );
+    }
 
     if ( $self->has_schema_hook ) {
         my $h = $self->schema_hook;
@@ -125,6 +209,28 @@ has storage => (
 
 sub _build_storage { shift->schema->storage }
 
+has for_update => (
+    isa => "Bool",
+    is  => "ro",
+    default => 1,
+);
+
+has _for_update => (
+    isa => "Bool",
+    is  => "ro",
+    lazy_build => 1,
+);
+
+sub _build__for_update {
+    my $self = shift;
+
+    return (
+        $self->for_update
+            and
+        $self->storage->sqlt_type =~ /^(?:MySQL|Oracle|PostgreSQL)$/
+    );
+}
+
 has columns => (
     isa => ArrayRef[ValidColumnName|HashRef],
     is  => "ro",
@@ -140,7 +246,7 @@ has _columns => (
 sub _build__columns {
     my $self = shift;
 
-    my $rs = $self->schema->source("entries");
+    my $rs = $self->schema->source( $self->kiokudb_entries_source_name );
 
     my @user_cols = grep { not exists $reserved_cols{$_} } $rs->columns;
 
@@ -187,6 +293,28 @@ sub _build_sql_abstract {
     SQL::Abstract->new;
 }
 
+sub register_handle {
+    my ( $self, $kiokudb ) = @_;
+
+    $self->schema->_kiokudb_handle($kiokudb);
+}
+
+sub default_typemap {
+    KiokuDB::TypeMap->new(
+        isa_entries => {
+            # redirect to schema row
+            'DBIx::Class::Row'                => KiokuDB::TypeMap::Entry::DBIC::Row->new,
+
+            # actual serialization
+            'DBIx::Class::ResultSet'          => KiokuDB::TypeMap::Entry::DBIC::ResultSet->new,
+
+            # fake, the entries never get written to the db
+            'DBIx::Class::ResultSourceHandle' => KiokuDB::TypeMap::Entry::DBIC::ResultSourceHandle->new,
+            'DBIx::Class::Schema'             => KiokuDB::TypeMap::Entry::DBIC::Schema->new,
+        },
+    );
+}
+
 sub insert {
     my ( $self, @entries ) = @_;
 
@@ -214,7 +342,7 @@ sub insert {
 sub entries_to_rows {
     my ( $self, @entries ) = @_;
 
-    my ( %insert, %update );
+    my ( %insert, %update, @dbic );
 
     foreach my $t ( \%insert, \%update ) {
         foreach my $col ( @{ $self->_ordered_columns } ) {
@@ -223,12 +351,20 @@ sub entries_to_rows {
     }
 
     foreach my $entry ( @entries ) {
-        my $targ = $entry->prev ? \%update : \%insert;
+        my $id = $entry->id;
 
-        my $row = $self->entry_to_row($entry, $targ);
+        if ( $id =~ /^dbic:schema:/ ) {
+            next;
+        } elsif ( $id =~ /^dbic:row:/ ) {
+            push @dbic, $entry->data;
+        } else {
+            my $targ = $entry->prev ? \%update : \%insert;
+
+            my $row = $self->entry_to_row($entry, $targ);
+        }
     }
 
-    return \( %insert, %update );
+    return \( %insert, %update, @dbic );
 }
 
 sub entry_to_row {
@@ -263,7 +399,7 @@ sub entry_to_row {
 }
 
 sub insert_rows {
-    my ( $self, $insert, $update ) = @_;
+    my ( $self, $insert, $update, $dbic ) = @_;
 
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
@@ -305,9 +441,17 @@ sub insert_rows {
 
             $sth->finish;
         }
+
+        $_->insert_or_update for @$dbic;
     });
 }
 
+sub prepare_select {
+    my ( $self, $dbh, $stmt ) = @_;
+
+    $dbh->prepare_cached($stmt . ( $self->_for_update ? " FOR UPDATE" : "" ));
+}
+
 sub prepare_insert {
     my ( $self, $dbh ) = @_;
 
@@ -349,39 +493,123 @@ sub update_index {
 }
 
 sub get {
-    my ( $self, @ids ) = @_;
+    my ( $self, @rows_and_ids ) = @_;
 
     my %entries;
 
-    $self->dbh_do(sub {
-        my ( $storage, $dbh ) = @_;
+    my ( @rows, @ids );
 
-        my $sth;
-
-        if ( @ids ) {
-            $sth = $dbh->prepare_cached("SELECT id, data FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
-            $sth->execute(@ids);
+    for ( @rows_and_ids ) {
+        if ( /^dbic:schema/ ) {
+            # schema objects aren't actually written into the DB
+            # FIXME special case FSCK
+            $entries{$_} = KiokuDB::Entry->new(
+                id    => $_,
+                ( $_ eq 'dbic:schema'
+                    ? ( class => 'DBIx::Class::Schema', data => $self->schema )
+                    : ( class => 'DBIx::Class::ResultSourceHandle', data => undef )
+                )
+            );
+        } elsif ( /^dbic:row:/ ) {
+            push @rows, $_;
         } else {
-            $sth = $dbh->prepare_cached("SELECT id, data FROM entries");
-            $sth->execute;
+            push @ids, $_;
+        }
+    }
+
+    if ( @ids ) {
+        $self->dbh_do(sub {
+            my ( $storage, $dbh ) = @_;
+
+            my $sth;
+
+            $sth = $self->prepare_select($dbh, "SELECT id, data FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
+            $sth->execute(@ids);
+
+            $sth->bind_columns( \my ( $id, $data ) );
+
+            # not actually necessary but i'm keeping it around for reference:
+            #my ( $id, $data );
+            #use DBD::Pg qw(PG_BYTEA);
+            #$sth->bind_col(1, \$id);
+            #$sth->bind_col(2, \$data, { pg_type => PG_BYTEA });
+
+            while ( $sth->fetch ) {
+                $entries{$id} = $data;
+            }
+        });
+    }
+
+    if ( @rows ) {
+        my $schema = $self->schema;
+        my $json = $self->json;
+
+        my ( %keys, %ids );
+
+        foreach my $id ( @rows ) {
+            my ( $rs_name, @key ) = @{ $json->decode(substr($id,length('dbic:row:'))) };
+
+            if ( @key > 2 ) {
+                # multi column primary keys need 'find'
+                my $obj = $schema->resultset($rs_name)->find(@key);
+
+                $entries{$id} = KiokuDB::Entry->new(
+                    id    => $id,
+                    class => ref($obj),
+                    data  => $obj,
+                );
+            } else {
+                # for other objects we queue up IDs for a single SELECT
+                push @{ $keys{$rs_name} ||= [] }, $key[0];
+                push @{ $ids{$rs_name}  ||= [] }, $id;
+            }
         }
 
-        $sth->bind_columns( \my ( $id, $data ) );
+        foreach my $rs_name ( keys %keys ) {
+            my $rs = $schema->resultset($rs_name);
 
-        # not actually necessary but i'm keeping it around for reference:
-        #my ( $id, $data );
-        #use DBD::Pg qw(PG_BYTEA);
-        #$sth->bind_col(1, \$id);
-        #$sth->bind_col(2, \$data, { pg_type => PG_BYTEA });
+            my $ids = $ids{$rs_name};
 
-        while ( $sth->fetch ) {
-            $entries{$id} = $data;
+            my @objs;
+
+            if ( @$ids == 1 ) {
+                my $id = $ids->[0];
+
+                my $obj = $rs->find($keys{$rs_name}[0]) or return;
+
+                $entries{$id} = KiokuDB::Entry->new(
+                    id => $id,
+                    class => ref($obj),
+                    data => $obj,
+                );
+            } else {
+                my @pk = $rs->result_source->primary_columns;
+
+                my $keys = $keys{$rs_name};
+
+                my @objs = $rs->search({ $pk[0] => $keys })->all;
+
+                return if @objs != @$ids;
+
+                # this key lookup is because it's not returned in the same order
+                my %pk_to_id;
+                @pk_to_id{@{ $keys{$rs_name} }} = @$ids;
+
+                foreach my $obj ( @objs ) {
+                    my $id = $pk_to_id{$obj->id};
+                    $entries{$id} = KiokuDB::Entry->new(
+                        id    => $id,
+                        class => ref($obj),
+                        data  => $obj,
+                    );
+                }
+            }
         }
-    });
+    }
 
-    return if @ids != keys %entries; # ->rows only works after we're done
+    return if @rows_and_ids != keys %entries; # ->rows only works after we're done
 
-    return map { $self->deserialize($_) } @entries{@ids};
+    map { ref($_) ? $_ : $self->deserialize($_) } @entries{@rows_and_ids};
 }
 
 sub delete {
@@ -415,7 +643,7 @@ sub exists {
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
-        my $sth = $dbh->prepare_cached("SELECT id FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
+        my $sth = $self->prepare_select($dbh, "SELECT id FROM entries WHERE id IN (" . join(", ", ('?') x @ids) . ")");
         $sth->execute(@ids);
 
         $sth->bind_columns( \( my $id ) );
@@ -552,13 +780,25 @@ sub insert_entry {
     die "Insertion to the GIN index is handled implicitly";
 }
 
+sub tables_exist {
+    my $self = shift;
+
+    $self->dbh_do(sub {
+        my ( $storage, $dbh ) = @_;
+
+        my $filter = ( $self->storage->sqlt_type eq 'SQLite' ? '%' : '' );
+
+        return ( @{ $dbh->table_info($filter, $filter, 'entries', 'TABLE')->fetchall_arrayref } > 0 );
+    });
+}
+
 sub create_tables {
     my $self = shift;
 
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
-        unless ( @{ $dbh->table_info('', '', 'entries', 'TABLE')->fetchall_arrayref } ) {
+        unless ( $self->tables_exist ) {
             $self->deploy({ producer_args => { mysql_version => 4.1 } });
         }
     });
@@ -644,6 +884,16 @@ will be C<NULL>.
 These columns are only used for lookup purposes, only C<data> is consulted when
 loading entries.
 
+=head1 DBIC INTEGRATION
+
+This backend is layered on top of L<DBIx::Class::Storage::DBI> and reused
+L<DBIx::Class::Schema> for DDL.
+
+Because of this objects from a L<DBIx::Class::Schema> can refer to objects in
+the KiokuDB entries table, and vice versa.
+
+For more details see L<DBIx::Class::Schema::KiokuDB>.
+
 =head1 SUPPORTED DATABASES
 
 This driver has been tested with MySQL 5 (4.1 should be the minimal supported
@@ -683,7 +933,7 @@ SQLite provides serializable isolation by default.
 
 L<http://www.sqlite.org/pragma.html#pragma_read_uncommitted>
 
-=head2 MySQL
+=head3 MySQL
 
 MySQL provides read committed isolation by default.
 
@@ -692,7 +942,7 @@ C<transaction-isolation> global variable,
 
 L<http://dev.mysql.com/doc/refman/5.1/en/set-transaction.html#isolevel_serializable>
 
-=head2 PostgreSQL
+=head3 PostgreSQL
 
 PostgreSQL provides read committed isolation by default.
 
@@ -765,6 +1015,50 @@ argument just before connecting.
 If you need to modify the schema in some way (adding indexes or constraints)
 this is where it should be done.
 
+=item for_update
+
+If true (the defaults), will cause all select statement to be issued with a
+C<FOR UPDATE> modifier on MySQL, Postgres and Oracle.
+
+This is highly reccomended because these database provide low isolation
+guarantees as configured out the box, and highly interlinked graph databases
+are much more susceptible to corruption because of lack of transcational
+isolation than normalized relational databases.
+
+=item sqlite_sync_mode
+
+If this attribute is set and the underlying database is SQLite, then
+C<PRAGMA syncrhonous=...> will be issued with this value.
+
+Can be C<OFF>, C<NORMAL> or C<FULL> (SQLite's default), or 0, 1, or 2.
+
+See L<http://www.sqlite.org/pragma.html#pragma_synchronous>.
+
+=item mysql_strict
+
+If true (the default), sets MySQL's strict mode.
+
+This is B<HIGHLY> reccomended, or you may enjoy some of MySQL's more
+interesting features, like automatic data loss when the columns are too narrow.
+
+See L<http://dev.mysql.com/doc/refman/5.0/en/server-sql-mode.html> and
+L<DBIx::Class::Storage::DBI::mysql> for more details.
+
+=item on_connect_call
+
+See L<DBIx::Class::Storage::DBI>.
+
+This attribute is constructed based on the values of C<mysql_version> and
+C<sqlite_sync_mode>, but may be overridden if you need more control.
+
+=item dbic_attrs
+
+See L<DBIx::Class::Storage::DBI>.
+
+Defaults to
+
+    { on_connect_call => $self->on_connect_call }
+
 =back
 
 =head1 METHODS
@@ -790,6 +1084,23 @@ Drops the C<entries> and C<gin_index> tables.
 
 =back
 
+=head1 TROUBLESHOOTING
+
+=head2 I get C<unexpected end of string while parsing JSON string>
+
+You are problably using MySQL, which comes with a helpful data compression
+feature: when your serialized objects are larger than the maximum size of a
+C<BLOB> column MySQL will simply shorten it for you.
+
+Why C<BLOB> defaults to 64k, and how on earth someone would consider silent
+data truncation a sane default I could never fathom, but nevertheless MySQL
+does allow you to disable this by setting the "strict" SQL mode in the
+configuration.
+
+To resolve the actual problem (though this obviously won't repair your lost
+data), alter the entries table so that the C<data> column uses the nonstandard
+C<LONGBLOB> datatype.
+
 =head1 VERSION CONTROL
 
 L<http://github.com/nothingmuch/kiokudb-backend-dbi>
